use std::sync::{Arc, Mutex};

use mls_rs::GroupStateStorage;
use mls_rs_core::group::{EpochRecord, GroupState};
use rusqlite::{Connection, OptionalExtension};

#[derive(Debug, Clone)]
/// SQLite Storage for MLS group states.
pub struct MlsCliGroupStateStorage {
    connection: Arc<Mutex<Connection>>,
    max_epoch_retention: u64,
}

impl MlsCliGroupStateStorage {
    pub fn new(connection: Arc<Mutex<Connection>>) -> MlsCliGroupStateStorage {
        MlsCliGroupStateStorage {
            connection,
            max_epoch_retention: 3,
        }
    }

    /// List all the group ids for groups that are stored.
    pub fn group_ids(&self) -> anyhow::Result<Vec<Vec<u8>>> {
        let connection = self.connection.lock().unwrap();
        let mut statement = connection.prepare("SELECT group_id FROM mls_group")?;
        let res = statement
            .query_map([], |row| row.get::<_, Vec<u8>>(0))?
            .try_fold(Vec::new(), |mut ids, id| {
                ids.push(id?);
                Ok::<_, anyhow::Error>(ids)
            })?;

        Ok(res)
    }

    /// Delete a group from storage.
    pub fn delete_group(&self, group_id: &[u8]) -> anyhow::Result<()> {
        let connection = self.connection.lock().unwrap();

        connection
            .execute(
                "DELETE FROM mls_group WHERE group_id = ?",
                rusqlite::params![group_id],
            )
            .map(|_| ())?;
        Ok(())
    }

    pub fn max_epoch_retention(&self) -> u64 {
        self.max_epoch_retention
    }

    fn get_snapshot_data(&self, group_id: &[u8]) -> anyhow::Result<Option<Vec<u8>>> {
        let connection = self.connection.lock().unwrap();

        connection
            .query_row(
                "SELECT snapshot FROM mls_group where group_id = ?",
                [group_id],
                |row| row.get::<_, Vec<u8>>(0),
            )
            .optional()
            .map_err(anyhow::Error::from)
    }

    fn get_epoch_data(&self, group_id: &[u8], epoch_id: u64) -> anyhow::Result<Option<Vec<u8>>> {
        let connection = self.connection.lock().unwrap();

        let res = connection
            .query_row(
                "SELECT epoch_data FROM epoch where group_id = ? AND epoch_id = ?",
                rusqlite::params![group_id, epoch_id],
                |row| row.get::<_, Vec<u8>>(0),
            )
            .optional()?;
        Ok(res)
    }

    fn max_epoch_id(&self, group_id: &[u8]) -> anyhow::Result<Option<u64>> {
        let connection = self.connection.lock().unwrap();

        connection
            .query_row(
                "SELECT MAX(epoch_id) FROM epoch WHERE group_id = ?",
                rusqlite::params![group_id],
                |row| row.get::<_, Option<u64>>(0),
            )
            .map_err(anyhow::Error::from)
    }

    fn update_group_state(
        &self,
        group_id: &[u8],
        group_snapshot: Vec<u8>,
        inserts: Vec<EpochRecord>,
        updates: Vec<EpochRecord>,
    ) -> anyhow::Result<()> {
        let mut max_epoch_id = None;

        let mut connection = self.connection.lock().unwrap();
        let transaction = connection.transaction()?;

        // Upsert into the group table to set the most recent snapshot
        transaction.execute(
            "INSERT INTO mls_group (group_id, snapshot) VALUES (?, ?) ON CONFLICT(group_id) DO UPDATE SET snapshot=excluded.snapshot",
            rusqlite::params![group_id, group_snapshot],
        )?;

        // Insert new epochs as needed
        for epoch in inserts {
            max_epoch_id = Some(epoch.id);

            transaction
                .execute(
                    "INSERT INTO epoch (group_id, epoch_id, epoch_data) VALUES (?, ?, ?)",
                    rusqlite::params![group_id, epoch.id, epoch.data],
                )
                .map(|_| ())?;
        }

        // Update existing epochs as needed
        for update in updates {
            max_epoch_id = Some(update.id);
            transaction
                .execute(
                    "UPDATE epoch SET epoch_data = ? WHERE group_id = ? AND epoch_id = ?",
                    rusqlite::params![update.data, group_id, update.id],
                )
                .map(|_| ())?;
        }

        // Delete old epochs as needed
        if let Some(max_epoch_id) = max_epoch_id {
            if max_epoch_id >= self.max_epoch_retention {
                let delete_under = max_epoch_id - self.max_epoch_retention;

                transaction.execute(
                    "DELETE FROM epoch WHERE group_id = ? AND epoch_id <= ?",
                    rusqlite::params![group_id, delete_under],
                )?;
            }
        }

        // Execute the full transaction
        transaction.commit()?;
        Ok(())
    }
}

impl GroupStateStorage for MlsCliGroupStateStorage {
    type Error = super::error::MlsCLIDBError;

    fn write(
        &mut self,
        state: GroupState,
        inserts: Vec<EpochRecord>,
        updates: Vec<EpochRecord>,
    ) -> Result<(), Self::Error> {
        let group_id = state.id;
        let snapshot_data = state.data;

        self.update_group_state(&group_id, snapshot_data, inserts, updates)
            .map_err(From::from)
    }

    fn state(&self, group_id: &[u8]) -> Result<Option<Vec<u8>>, Self::Error> {
        self.get_snapshot_data(group_id).map_err(From::from)
    }

    fn max_epoch_id(&self, group_id: &[u8]) -> Result<Option<u64>, Self::Error> {
        self.max_epoch_id(group_id).map_err(From::from)
    }

    fn epoch(&self, group_id: &[u8], epoch_id: u64) -> Result<Option<Vec<u8>>, Self::Error> {
        self.get_epoch_data(group_id, epoch_id).map_err(From::from)
    }
}
