use mls_rs_core::{
    key_package::{KeyPackageData, KeyPackageStorage},
    mls_rs_codec::{MlsDecode, MlsEncode},
    time::MlsTime,
};
use rusqlite::{Connection, OptionalExtension, params};
use std::sync::{Arc, Mutex};

#[derive(Debug, Clone)]
/// SQLite storage for MLS Key Packages.
pub struct MlsCliKeyPackageStorage {
    connection: Arc<Mutex<Connection>>,
}

impl MlsCliKeyPackageStorage {
    pub fn new(connection: Arc<Mutex<Connection>>) -> MlsCliKeyPackageStorage {
        MlsCliKeyPackageStorage { connection }
    }

    fn insert(&mut self, id: &[u8], key_package: KeyPackageData) -> anyhow::Result<()> {
        let connection = self.connection.lock().unwrap();

        connection
            .execute(
                "INSERT INTO key_package (id, expiration, data) VALUES (?,?,?)",
                params![id, key_package.expiration, key_package.mls_encode_to_vec()?],
            )
            .map(|_| ())
            .map_err(anyhow::Error::from)
    }

    fn get(&self, id: &[u8]) -> anyhow::Result<Option<KeyPackageData>> {
        let connection = self.connection.lock().unwrap();

        connection
            .query_row(
                "SELECT data FROM key_package WHERE id = ?",
                params![id],
                |row| {
                    Ok(
                        KeyPackageData::mls_decode(&mut row.get::<_, Vec<u8>>(0)?.as_slice())
                            .unwrap(),
                    )
                },
            )
            .optional()
            .map_err(anyhow::Error::from)
    }

    /// Delete a specific key package from storage based on it's id.
    pub fn delete(&self, id: &[u8]) -> anyhow::Result<()> {
        let connection = self.connection.lock().unwrap();

        connection
            .execute("DELETE FROM key_package where id = ?", params![id])
            .map(|_| ())
            .map_err(anyhow::Error::from)
    }

    /// Delete key packages that are expired based on the current system clock time.
    pub fn delete_expired(&self) -> anyhow::Result<()> {
        self.delete_expired_by_time(MlsTime::now().seconds_since_epoch())
    }

    /// Delete key packages that are expired based on an application provided time in seconds since
    /// unix epoch.
    pub fn delete_expired_by_time(&self, time: u64) -> anyhow::Result<()> {
        let connection = self.connection.lock().unwrap();

        connection
            .execute(
                "DELETE FROM key_package where expiration < ?",
                params![time],
            )
            .map(|_| ())
            .map_err(anyhow::Error::from)
    }

    /// Total number of key packages held in storage.
    pub fn count(&self) -> anyhow::Result<usize> {
        let connection = self.connection.lock().unwrap();

        connection
            .query_row("SELECT count(*) FROM key_package", params![], |row| {
                row.get(0)
            })
            .map_err(anyhow::Error::from)
    }

    pub fn count_at_time(&self, time: u64) -> anyhow::Result<usize> {
        self.delete_expired()?;
        let connection = self.connection.lock().unwrap();

        connection
            .query_row(
                "SELECT count(*) FROM key_package where expiration >= ?",
                params![time],
                |row| row.get(0),
            )
            .map_err(anyhow::Error::from)
    }
}

impl KeyPackageStorage for MlsCliKeyPackageStorage {
    type Error = super::error::MlsCLIDBError;

    fn insert(&mut self, id: Vec<u8>, pkg: KeyPackageData) -> Result<(), Self::Error> {
        self.insert(id.as_slice(), pkg).map_err(From::from)
    }

    fn get(&self, id: &[u8]) -> Result<Option<KeyPackageData>, Self::Error> {
        self.get(id).map_err(From::from)
    }

    fn delete(&mut self, id: &[u8]) -> Result<(), Self::Error> {
        (*self).delete(id).map_err(From::from)
    }
}
