use std::sync::{Arc, Mutex};

use chrono::{DateTime, Utc};
use rusqlite::Connection;

#[derive(Debug, Clone)]
pub struct MessageStorage {
    pub connection: Arc<Mutex<Connection>>,
}

pub struct MessageDTO {
    pub group_id: Vec<u8>,
    pub sender_id: Vec<u8>,
    pub data: Vec<u8>,
    pub created_at: DateTime<Utc>,
}

impl MessageStorage {
    pub fn new(connection: Arc<Mutex<Connection>>) -> Self {
        MessageStorage { connection }
    }

    pub fn insert_message(
        &self,
        chat_id: &[u8],
        sender_id: &[u8],
        message: &[u8],
        epoch_id: u64,
    ) -> anyhow::Result<()> {
        let conn = self
            .connection
            .lock()
            .expect("Connection mutext is poisoned. It's a critical error. Exiting.");
        conn.execute(
            "INSERT INTO mls_cli_messages(
                group_id, 
                sender_id,
                data,
                epoch_id
            ) VALUES (
                :group_id, 
                :sender_id,
                :data,
                :epoch_id
            )",
            rusqlite::named_params![
                ":group_id": chat_id,
                ":sender_id": sender_id,
                ":data": message,
                ":epoch_id": epoch_id
            ],
        )?;
        Ok(())
    }

    #[allow(dead_code)]
    pub fn get_messages(
        &self,
        group_id: &str,
        offset: i64,
        limit: i64,
    ) -> anyhow::Result<Vec<MessageDTO>> {
        let conn = self
            .connection
            .lock()
            .expect("Connection mutext is poisoned. It's a critical error. Exiting.");
        let mut stmt = conn.prepare(
            "SELECT 
                group_id, 
                sender_id,
                data,
                created_at
            FROM mls_cli_messages 
            WHERE group_id = :group_id
            ORDER BY created_at DESC
            LIMIT :limit
            OFFSET :offset",
        )?;
        let rows = stmt.query_map(
            rusqlite::named_params![
                ":group_id": group_id.as_bytes(),
                ":offset": offset,
                ":limit": limit
            ],
            |row| {
                let group_id = row.get(0)?;
                let sender_id = row.get(1)?;
                let data = row.get(2)?;
                let created_at = row.get(3)?;
                Ok(MessageDTO {
                    group_id,
                    sender_id,
                    data,
                    created_at,
                })
            },
        )?;
        let mut res = vec![];
        for row in rows {
            res.push(row?);
        }
        Ok(res)
    }
    #[allow(dead_code)]
    pub fn get_all_messages(&self, offset: i64, limit: i64) -> anyhow::Result<Vec<MessageDTO>> {
        let conn = self
            .connection
            .lock()
            .expect("Connection mutext is poisoned. It's a critical error. Exiting.");
        let mut stmt = conn.prepare(
            "SELECT 
                group_id, 
                sender_id,
                data,
                created_at
            FROM mls_cli_messages 
            ORDER BY created_at DESC 
            LIMIT :limit
            OFFSET :offset",
        )?;
        let rows = stmt.query_map(
            rusqlite::named_params![":offset": offset, ":limit": limit],
            |row| {
                let group_id = row.get(0)?;
                let sender_id = row.get(1)?;
                let data = row.get(2)?;
                let created_at = row.get(3)?;
                Ok(MessageDTO {
                    group_id,
                    sender_id,
                    data,
                    created_at,
                })
            },
        )?;
        let mut res = vec![];
        for row in rows {
            res.push(row?);
        }
        Ok(res)
    }

    pub fn message_sent_in_this_epoch(&self, group_id: &str, epoch_id: u64) -> anyhow::Result<u64> {
        let conn = self
            .connection
            .lock()
            .expect("Connection mutext is poisoned. It's a critical error. Exiting.");
        let mut stmt = conn.prepare(
            "SELECT 
                COUNT(*)
            FROM mls_cli_messages 
            WHERE group_id = :group_id
            AND epoch_id = :epoch_id",
        )?;
        let count = stmt.query_row(
            rusqlite::named_params![
                ":group_id": group_id.as_bytes(),
                ":epoch_id": epoch_id
            ],
            |row| row.get(0),
        )?;
        Ok(count)
    }
}
