use mls_rs_core::psk::{ExternalPskId, PreSharedKey, PreSharedKeyStorage};
use rusqlite::{Connection, OptionalExtension, params};
use std::{
    ops::Deref,
    sync::{Arc, Mutex},
};

#[derive(Debug, Clone)]
/// SQLite storage for MLS pre-shared keys.
pub struct MlsCliPreSharedKeyStorage {
    connection: Arc<Mutex<Connection>>,
}

impl MlsCliPreSharedKeyStorage {
    pub fn new(connection: Arc<Mutex<Connection>>) -> Self {
        MlsCliPreSharedKeyStorage { connection }
    }

    /// Insert a pre-shared key into storage.
    pub fn insert(&self, psk_id: &[u8], psk: &PreSharedKey) -> anyhow::Result<()> {
        let connection = self.connection.lock().unwrap();

        // Upsert into the database
        connection
            .execute(
                "INSERT INTO psk (psk_id, data) VALUES (?,?) ON CONFLICT(psk_id) DO UPDATE SET data=excluded.data",
                params![psk_id, psk.deref()],
            )
            .map(|_| ()).map_err(anyhow::Error::from)
    }

    /// Get a pre-shared key from storage based on a unique id.
    pub fn get(&self, psk_id: &[u8]) -> anyhow::Result<Option<PreSharedKey>> {
        let connection = self.connection.lock().unwrap();

        connection
            .query_row(
                "SELECT data FROM psk WHERE psk_id = ?",
                params![psk_id],
                |row| Ok(PreSharedKey::new(row.get(0)?)),
            )
            .optional()
            .map_err(anyhow::Error::from)
    }

    /// Delete a pre-shared key from storage based on a unique id.
    pub fn delete(&self, psk_id: &[u8]) -> anyhow::Result<()> {
        let connection = self.connection.lock().unwrap();

        connection
            .execute("DELETE FROM psk WHERE psk_id = ?", params![psk_id])
            .map(|_| ())
            .map_err(anyhow::Error::from)
    }
}

impl PreSharedKeyStorage for MlsCliPreSharedKeyStorage {
    type Error = super::error::MlsCLIDBError;

    fn get(&self, id: &ExternalPskId) -> Result<Option<PreSharedKey>, Self::Error> {
        self.get(id).map_err(From::from)
    }
}
