use std::collections::HashMap;
use std::sync::Arc;

use crate::context::ClientContext;
use mls_rs::MlsMessage;
use mls_rs::group::ReceivedMessage;
use mls_rs::time::MlsTime;
use tokio::sync::RwLock;

pub fn process_updates(ctx: Arc<RwLock<ClientContext>>) {
    // Messages not related to any group
    let mut general_messages: Vec<MlsMessage> = vec![];
    // Messages related to some groups
    let mut grouped_updates: HashMap<Vec<u8>, Vec<MlsMessage>> = HashMap::default();
    let ctx = ctx.read();
    let Some(mls_ctx) = &ctx.mls else { return };
    let updates = ctx.ds_client.get_updates().await;

    match updates {
        Ok(updates) => {
            for update in updates {
                if let Some(group_id) = update.group_id().map(|id| id.to_vec()) {
                    grouped_updates.entry(group_id).or_default().push(update);
                    continue;
                } else {
                    general_messages.push(update);
                }
            }
        }
        Err(err) => {
            maybe_printer_print(
                format!("Cannot receive updates! Reason: {}", err.to_string().red()),
                printer,
            );
        }
    };

    if !general_messages.is_empty() {
        for msg in &general_messages {
            if msg.wire_format() == mls_rs::WireFormat::Welcome {
                match mls_ctx.mls_client.join_group(None, msg) {
                    Ok((mut group, _)) => {
                        group.write_to_storage().ok();
                    }
                    Err(err) => {
                        maybe_printer_print(
                            format!("Cannot join group! Reason: {}", err.to_string().red()),
                            printer,
                        );
                    }
                }
            }
        }
    }
    if !grouped_updates.is_empty() {
        for (group_id, messages) in &grouped_updates {
            let Ok(mut group) = mls_ctx.mls_client.load_group(group_id) else {
                maybe_printer_print(
                    format!(
                        "Cannot load group with id: {:?}",
                        String::from_utf8_lossy(group_id).red().to_string()
                    ),
                    printer,
                );
                continue;
            };

            for message in messages {
                match group.process_incoming_message_with_time(message.clone(), MlsTime::now()) {
                    Ok(msg) => match msg {
                        ReceivedMessage::ApplicationMessage(app_msg) => {
                            let Some(member) = group.member_at_index(app_msg.sender_index) else {
                                maybe_printer_print(
                                    format!(
                                        "Cannot find member with index: {}",
                                        app_msg.sender_index
                                    ),
                                    printer,
                                );
                                continue;
                            };
                            let sender_name = member
                                .signing_identity
                                .credential
                                .as_basic()
                                .unwrap()
                                .identifier
                                .as_slice();
                            let content = String::from_utf8_lossy(app_msg.data());
                            let group_name = String::from_utf8_lossy(group_id);
                            ctx.message_storage
                                .insert_message(
                                    group_id,
                                    sender_name,
                                    app_msg.data(),
                                    group.context().epoch(),
                                )
                                .ok();
                            eprintln!(
                                "Received a message. group: `{}`; content: `{}`",
                                group_name.yellow(),
                                content.green(),
                            );
                        }
                        ReceivedMessage::Proposal(proposal) => {
                            let commit = group
                                .commit_builder()
                                .raw_proposal(proposal.proposal)
                                .build();
                            match commit {
                                Ok(cmt) => {
                                    group.apply_pending_commit().ok();
                                    ctx.ds_client.post_commit(&cmt).await.ok();
                                }
                                Err(err) => {
                                    maybe_printer_print(
                                        format!(
                                            "Cannot create commit! Reason: {}",
                                            err.to_string().red()
                                        ),
                                        printer,
                                    );
                                }
                            }
                        }
                        _ => {}
                    },
                    Err(err) => match err {
                        // Ignore messages from self
                        mls_rs::error::MlsError::CantProcessMessageFromSelf => {}
                        _ => {
                            maybe_printer_print(
                                format!(
                                    "Cannot process message! Reason: {}",
                                    err.to_string().red()
                                ),
                                printer,
                            );
                        }
                    },
                }
            }
            group.write_to_storage().ok();
        }
    }
    Vec::clear(&mut general_messages);
    HashMap::clear(&mut grouped_updates);
}

pub async fn listen_to_updates(
    ctx: Arc<RwLock<ClientContext>>,
    printer: ExternalPrinter<String>,
) -> ! {
    loop {
        tokio::time::sleep(std::time::Duration::from_secs(1)).await;
        process_updates(ctx.clone(), Some(&printer)).await;
    }
}
