use std::{fs::OpenOptions, path::Path};

use axum::{
    Extension,
    extract::{State, ws::Message},
    http::HeaderMap,
    response::IntoResponse,
};
use clap::Parser;
use futures::SinkExt;
use futures::StreamExt;
use mls_rs::mls_rs_codec::MlsDecode;
use sqlx::sqlite::{SqlitePool, SqlitePoolOptions};
use state::MlcState;
use tokio::net::TcpListener;
use tracing::level_filters::LevelFilter;
use tracing_subscriber::{EnvFilter, FmtSubscriber};

mod args;
mod daos;
mod routes;
mod services;
mod state;

async fn start() -> anyhow::Result<()> {
    let config = args::MlcConfig::parse();
    // Get the database URL from environment variables
    {
        if let Some(parent) = Path::new(&config.db_path).parent() {
            std::fs::create_dir_all(&parent)?;
        }
        OpenOptions::new()
            .create(true)
            .write(true)
            .open(&config.db_path)?;
    }
    tracing::info!("Database path: {}", config.db_path);
    let sqlite_pool = SqlitePoolOptions::new()
        .max_connections(10)
        .min_connections(4)
        .max_lifetime(None)
        .connect(&format!("sqlite:{}", config.db_path))
        .await
        .expect("Failed to connect to database");

    // Run migrations
    sqlx::migrate!("./migrations")
        .run(&sqlite_pool)
        .await
        .expect("Failed to run migrations");

    let state = state::MlcState {
        sqlite_pool,
        ws_manager: services::WSManager::default(),
        config: config.clone(),
    };

    let app = axum::Router::new()
        .route("/ws", axum::routing::get(ws_handler))
        .with_state(state);

    let listener = TcpListener::bind((config.server_host, config.server_port)).await?;
    tracing::info!("Listening on: {}", listener.local_addr()?);
    axum::serve(listener, app).await?;
    Ok(())
}

async fn ws_handler(
    ws: axum::extract::ws::WebSocketUpgrade,
    headers: HeaderMap,
    State(db_pool): State<MlcState>,
) -> impl IntoResponse {
    ws.on_upgrade(move |socket| {
        let username = headers.get("username").unwrap().to_str().unwrap();
        let (ws_writer, ws_reader) = socket.split();
        handle_messages(ws_writer, ws_reader, username.to_string(), db_pool)
    })
}

async fn handle_messages<W, R>(
    mut ws_writer: W,
    mut ws_reader: R,
    username: String,
    state: MlcState,
) where
    W: futures::Sink<Message> + Unpin,
    R: futures::Stream<Item = Result<Message, axum::Error>> + Unpin,
    <W as futures::Sink<Message>>::Error: std::fmt::Display,
{
    let (messages_tx, mut messages_rx) =
        tokio::sync::mpsc::channel::<String>(state.config.per_user_msg_buffer);
    let ws_id = state.ws_manager.add_client(&username, messages_tx).await;
    tracing::Span::current().record("ws_id", ws_id);
    let mut messages = vec![];
    loop {
        tokio::select! {
            msg = messages_rx.recv() => {
                if let Some(msg) = msg {
                    ws_writer.send(Message::Text(msg.into())).await.ok();
                }else{
                    tracing::debug!("Messages channel has been closed");
                    break;
                }
            },
            msg = ws_reader.next() => {
                let Some(msg_res) = msg else {
                    tracing::debug!("Websocket connection has been closed");
                    break;
                };
                let Ok(msg) = msg_res else {
                    tracing::warn!("Cannot receive message from client: {}", msg_res.err().unwrap());
                    break;
                };
                match msg {
                    Message::Close(_) => {
                        tracing::debug!("Client has closed the connection");
                        break;
                    }
                    // Message::Ping(bytes) => {ws_writer.send(Message::Pong(bytes)).await.ok();},
                    Message::Text(utf8_bytes) => {
                        tracing::info!("Received message from ws: {:?}", utf8_bytes);
                        if let Err(err) = ws_writer.send(Message::Text(utf8_bytes)).await {
                            tracing::error!("Failed to send message back: {}", err);
                            break;
                        }
                    },
                    Message::Binary(bytes) => {
                        // bytes.
                        let mut raw_bytes = bytes.as_ref();
                        while raw_bytes.len() > 0 {
                            if let Ok(msg) = mls_rs::MlsMessage::mls_decode(&mut raw_bytes){
                                messages.push(msg);
                            } else {break}
                        }
                    },
                    _ => {}
                }
            }
        }
        println!("{:?}", messages);
        messages.truncate(0);
    }
}

fn main() -> anyhow::Result<()> {
    // Load environment variables from .env file
    dotenv::dotenv().ok();
    FmtSubscriber::builder()
        .with_env_filter(
            EnvFilter::builder()
                .with_default_directive(LevelFilter::INFO.into())
                .from_env_lossy(),
        )
        .with_writer(std::io::stderr)
        .init();

    // Start the HTTP server
    let rt = tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .build()?;

    rt.block_on(start())
}
