use std::sync::Arc;

use rustc_hash::FxHashMap;
use tokio::sync::{RwLock, mpsc::Sender};

#[derive(Clone, Debug, Default)]
pub struct WSManager {
    connected_clients: Arc<RwLock<FxHashMap<String, Vec<Sender<String>>>>>,
}

impl WSManager {
    #[tracing::instrument(level = "debug", skip(self, sender))]
    pub async fn add_client(&self, user_id: &str, sender: Sender<String>) -> usize {
        let mut connected_clients = self.connected_clients.write().await;
        let senders = connected_clients
            .entry(user_id.to_string())
            .or_insert_with(Vec::new);
        senders.push(sender);
        // Minus one, because we use zero-based indexing.
        senders.len() - 1
    }

    #[tracing::instrument(level = "debug", skip(self))]
    pub async fn send_message(&self, user_id: &str, message: &str) -> bool {
        let mut dead_sockets = vec![];
        let mut message_sent = false;
        let user_id_str = user_id.to_string();
        // Here we are defining a scope to drop the read lock as soon as we are done with it.
        // Because later we will need to acquire a write lock on the same data.
        // If we won't drop the read lock, we will get a deadlock.
        {
            let connected_clients = self.connected_clients.read().await;
            if let Some(senders) = connected_clients.get(&user_id_str) {
                for (id, sender) in senders.iter().enumerate() {
                    if sender.send(message.to_string()).await.is_err() {
                        dead_sockets.push(id);
                    }
                }
                // The delivery was successful only if we have
                // sent the message to at least one socket.
                if dead_sockets.len() < senders.len() {
                    message_sent = true;
                }
            }
        }
        // If all sockets that are connected to the user are running fine, we can return.
        if dead_sockets.is_empty() {
            return message_sent;
        }

        self.remove_clients(user_id, dead_sockets).await;

        message_sent
    }

    #[tracing::instrument(level = "debug", skip(self))]
    pub async fn remove_clients(&self, user_id: &str, ws_ids: Vec<usize>) {
        let mut connected_clients = self.connected_clients.write().await;
        let mut senders_left = 0;
        if let Some(senders) = connected_clients.get_mut(&user_id.to_string()) {
            for id in ws_ids {
                if id >= senders.len() {
                    continue;
                }
                senders.swap_remove(id);
            }
            senders_left = senders.len();
        }
        if senders_left == 0 {
            connected_clients.remove(&user_id.to_string());
        }
    }
}

#[cfg(test)]
mod tests {
    #[tokio::test]
    async fn add_client() {
        let ws_manager = super::WSManager::default();
        let (sender, _) = tokio::sync::mpsc::channel(1);
        let uid = uuid::Uuid::new_v4().to_string();
        let client_id = ws_manager.add_client(&uid, sender).await;
        assert_eq!(client_id, 0);
        let connected_clients = ws_manager.connected_clients.read().await;
        let sockets = connected_clients.get(&uid.to_string()).unwrap();
        assert_eq!(sockets.len(), 1);
    }

    #[tokio::test]
    async fn sending_message() {
        let ws_manager = super::WSManager::default();
        let (sender, mut receiver) = tokio::sync::mpsc::channel(1);
        let uid = uuid::Uuid::new_v4().to_string();
        let client_id = ws_manager.add_client(&uid, sender).await;
        assert_eq!(client_id, 0);
        ws_manager.send_message(&uid, "test").await;
        let message = receiver.recv().await;
        assert_eq!(message, Some("test".to_string()));
    }

    #[tokio::test]
    async fn multiple_clients() {
        let ws_manager = super::WSManager::default();
        let mut receivers = vec![];
        let uid = uuid::Uuid::new_v4().to_string();
        let num_clients = 10;
        for i in 0..num_clients {
            let (sender, receiver) = tokio::sync::mpsc::channel(1);
            let client_id = ws_manager.add_client(&uid, sender).await;
            assert_eq!(client_id, i);
            receivers.push(receiver);
        }
        ws_manager.send_message(&uid, "test").await;
        for mut receiver in receivers {
            let message = receiver.recv().await;
            assert_eq!(message, Some("test".to_string()));
        }
    }

    #[tokio::test]
    async fn remove_client() {
        let ws_manager = super::WSManager::default();
        let (sender, mut receiver) = tokio::sync::mpsc::channel(1);
        let uid = uuid::Uuid::new_v4().to_string();
        let client_id = ws_manager.add_client(&uid, sender).await;
        assert_eq!(client_id, 0);
        ws_manager.remove_clients(&uid, vec![0, 1, 2]).await;
        ws_manager.send_message(&uid, "test").await;
        let message = receiver.recv().await;
        assert_eq!(message, None);
        let all_clients = ws_manager.connected_clients.read().await;
        assert_eq!(all_clients.len(), 0);
    }

    #[tokio::test]
    async fn removing_dead_client() {
        let ws_manager = super::WSManager::default();
        let (sender, mut receiver) = tokio::sync::mpsc::channel(1);
        let uid = uuid::Uuid::new_v4().to_string();
        let client_id = ws_manager.add_client(&uid, sender).await;
        receiver.close();
        assert_eq!(client_id, 0);
        let delivered = ws_manager.send_message(&uid, "test").await;
        assert_eq!(delivered, false);
        let all_clients = ws_manager.connected_clients.read().await;
        assert_eq!(all_clients.len(), 0);
    }
}
